#import "KinectPatch.h"

uint16_t t_gamma[2048];


void KinectPatchDepthBufferCallback(uint16_t *buf, int width, int height, void *context)
{
	KinectPatch *k = (KinectPatch *)context;
	[k updateDepthImageWithBuffer:buf];
}

void KinectPatchColorBufferCallback(uint8_t *buf, int width, int height, void *context)
{
	KinectPatch *k = (KinectPatch *)context;
	[k updateColorImageWithBuffer:buf];
}



@implementation KinectPatch

+(BOOL)allowsSubpatchesWithIdentifier:(id)identifier
{
	return NO;
}

+(QCPatchExecutionMode)executionModeWithIdentifier:(id)identifier
{
	// defaults to kQCPatchExecutionModeProcessor.
	// delete this method if using default.
	return kQCPatchExecutionModeProvider;
}

-(id)initWithIdentifier:(id)identifier
{
	if(self = [super initWithIdentifier:identifier])
	{
		[[self userInfo] setObject:@"Kineme Kinect Patch" forKey:@"name"];

		int i;
		for (i=0; i<2048; i++) {
			float v = i/2048.0;
			v = powf(v, 3)* 6;
			t_gamma[i] = v*6*256;
		}
	}
	return self;
}

-(void)enable:(QCOpenGLContext*)context
{
	[outputColorImage setImageValue:nil];
	[outputDepthImage setImageValue:nil];

	if(libusb_init(&uContext))
	{
		SLog(@"Failed to initialize libusb.");
		return;
	}

	uDevice = libusb_open_device_with_vid_pid(uContext, 0x45e, 0x2ae);
	if(!uDevice)
	{
		SLog(@"Could not open USB device.");
		return;
	}
	SLog(@"Opened USB device: %i",libusb_get_device_address(libusb_get_device(uDevice)));
	
	if(libusb_claim_interface(uDevice,0))
		SLog(@"Failed to claim interface.");

	glContext=context;
	cams_init(uDevice, KinectPatchDepthBufferCallback, KinectPatchColorBufferCallback, (void *)self);
}

-(void)disable:(QCOpenGLContext*)context
{
	if(uDevice)
	{
		SLog(@"Closing USB device.");
		cams_exit();
		if(libusb_release_interface(uDevice, 0))
			SLog(@"Failed to release interface.");
		uDevice=0;
	}
	if(uContext)
	{
		libusb_exit(uContext);
		uContext=0;
	}
}



- (void)updateDepthImageWithBuffer:(uint16_t *)buf
{
//	SLog();
	QCImagePixelBuffer *pb = [[glContext imageManager] createPixelBufferWithFormat: [QCPixelFormat pixelFormatARGB8]
																	  pixelsWide: 640
																	  pixelsHigh: 480
																		 options: nil];
	CGColorSpaceRef cs = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
	[pb beginUpdatePixels:TRUE colorSpace:cs];

	char *base = [pb baseAddress];
	int basestride = 640*4;
	if ([[pb pixelBufferRepresentation] isFlipped])
	{
//		SLog(@"Flipping.");
		base += 479*640*4;
		basestride *= -1;
	}

	// @@@ todo: make this not so horrifically inefficient, now that I've kinda figured out what's going on here.
	for(NSUInteger y=0;y<480;++y,base+=basestride)
	{
		for(NSUInteger x=0;x<640;++x)
		{
			uint16_t v = *(buf+y*640+x);

			// 0x7ff is apparently "I couldn't figure out what depth this pixel is"
			if(v == 0x7ff)
				base[x*4+0]= 0;
			else
			{
				base[x*4+0]= 255;

				base[x*4+1]=
				base[x*4+2]=
				base[x*4+3]= (t_gamma[v]>>3)&0xff;
			}
		}
	}


	[pb endUpdatePixels];
	CGColorSpaceRelease(cs);
	QCImage *image = [[QCImage alloc] initWithQCImageBuffer:pb options:nil]; 
	[pb release]; 
	[outputDepthImage setImageValue:image];
	[image release];
}
- (void)updateColorImageWithBuffer:(uint8_t *)buf
{
//	SLog();
	QCImagePixelBuffer *pb = [[glContext imageManager] createPixelBufferWithFormat: [QCPixelFormat pixelFormatRGB8]
																	  pixelsWide: 640
																	  pixelsHigh: 480
																		 options: nil];
	CGColorSpaceRef cs = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB); 
	[pb beginUpdatePixels:TRUE colorSpace:cs];

	if ([[pb pixelBufferRepresentation] isFlipped])
	{
//		SLog(@"Flipping.");
		char *base = [pb baseAddress] + 479*640*3;
		for(NSUInteger y=0;y<480;++y,base-=640*3)
			memcpy(base, buf+y*640*3, 640*3);
	}
	else
		memcpy([pb baseAddress], buf, 640*480*3);

	[pb endUpdatePixels];
	CGColorSpaceRelease(cs);
	QCImage *image = [[QCImage alloc] initWithQCImageBuffer:pb options:nil]; 
	[pb release]; 
	[outputColorImage setImageValue:image];
	[image release];
}

-(BOOL)execute:(QCOpenGLContext*)context time:(double)time arguments:(NSDictionary*)arguments
{
//	SLog();
	if(!uDevice)
		[outputConnected setBooleanValue:NO];
	else
	{
		[outputConnected setBooleanValue:YES];

		glContext=context;
//	[self updateDepthImageWithBuffer:malloc(640*480*3)];
//	[self updateColorImageWithBuffer:malloc(640*480*3)];
		struct timeval tv;
		tv.tv_sec = 0;
		tv.tv_usec = 0;
		libusb_handle_events_timeout(uContext,&tv);
	}
	return YES;
}

@end
